"""Slicer tester."""
from unittest import TestCase, main

import pandas as pd

from soam.workflow import Slicer


class TestSlicer(TestCase):
    """Creates the slicer tester object."""

    columns = ["date", "letter", "move", "opportunities", "impressions", "revenue"]
    values = [
        ["2019-09-01", "A", "down", 1000, 100, 20],
        ["2019-09-01", "B", "down", 500, 68, 30],
        ["2019-09-02", "A", "down", 400, 300, 52],
        ["2019-09-02", "B", "up", 300, 35, 6],
        ["2019-09-03", "A", "up", 200, 10, 40],
        ["2019-09-03", "B", "up", 700, 30, 9],
    ]
    df = pd.DataFrame(data=values, columns=columns)

    def _test_slices(self, dimensions, metrics, length, dfs):

        slicer_test = Slicer(dimensions=dimensions, metrics=metrics, ds_col="date")
        dataframes = slicer_test.run(self.df)

        self.assertEqual(len(dataframes), length)
        for df_returned, df_passed in zip(dataframes, dfs):
            df_returned.reset_index(inplace=True, drop=True)
            df_passed.reset_index(inplace=True, drop=True)
            pd.testing.assert_frame_equal(
                df_returned, df_passed, check_index_type=False, check_like=True
            )

    def test_slice_one_column(self):
        """Tests the slice on one column."""
        columns = ["date", "letter", "opportunities"]
        df1 = pd.DataFrame(
            columns=columns,
            data=[
                ["2019-09-01", "A", 1000],
                ["2019-09-02", "A", 400],
                ["2019-09-03", "A", 200],
            ],
        )
        df2 = pd.DataFrame(
            columns=columns,
            data=[
                ["2019-09-01", "B", 500],
                ["2019-09-02", "B", 300],
                ["2019-09-03", "B", 700],
            ],
        )
        dfs = [df1, df2]
        self._test_slices("letter", "opportunities", 2, dfs)

    def test_slice_two_column(self):
        """Tests the slice on two columns."""
        columns1 = ["date", "letter", "opportunities"]
        columns2 = ["date", "move", "opportunities"]
        df1 = pd.DataFrame(
            columns=columns1,
            data=[
                ["2019-09-01", "A", 1000],
                ["2019-09-02", "A", 400],
                ["2019-09-03", "A", 200],
            ],
        )
        df2 = pd.DataFrame(
            columns=columns1,
            data=[
                ["2019-09-01", "B", 500],
                ["2019-09-02", "B", 300],
                ["2019-09-03", "B", 700],
            ],
        )
        df3 = pd.DataFrame(
            columns=columns2,
            data=[
                ["2019-09-01", "down", 1000],
                ["2019-09-01", "down", 500],
                ["2019-09-02", "down", 400],
            ],
        )
        df4 = pd.DataFrame(
            columns=columns2,
            data=[
                ["2019-09-02", "up", 300],
                ["2019-09-03", "up", 200],
                ["2019-09-03", "up", 700],
            ],
        )
        dfs = [df1, df2, df3, df4]
        self._test_slices(["letter", "move"], "opportunities", 4, dfs)

    def test_slice_two_dimensions(self):
        """Tests the slice on two dimensions."""
        columns = ["date", "letter", "move", "opportunities"]
        df1 = pd.DataFrame(
            columns=columns,
            data=[["2019-09-01", "A", "down", 1000], ["2019-09-02", "A", "down", 400],],
        )

        df2 = pd.DataFrame(columns=columns, data=[["2019-09-03", "A", "up", 200],])
        df3 = pd.DataFrame(columns=columns, data=[["2019-09-01", "B", "down", 500],])
        df4 = pd.DataFrame(
            columns=columns,
            data=[["2019-09-02", "B", "up", 300], ["2019-09-03", "B", "up", 700],],
        )
        dfs = [df1, df2, df3, df4]
        self._test_slices([["letter", "move"]], "opportunities", 4, dfs)

    def test_slice_two_metrics(self):
        """Tests the slice on two metrics."""
        columns1 = ["date", "letter", "opportunities"]
        columns2 = ["date", "letter", "impressions"]
        df1 = pd.DataFrame(
            columns=columns1,
            data=[
                ["2019-09-01", "A", 1000],
                ["2019-09-02", "A", 400],
                ["2019-09-03", "A", 200],
            ],
        )
        df3 = pd.DataFrame(
            columns=columns1,
            data=[
                ["2019-09-01", "B", 500],
                ["2019-09-02", "B", 300],
                ["2019-09-03", "B", 700],
            ],
        )
        df2 = pd.DataFrame(
            columns=columns2,
            data=[
                ["2019-09-01", "A", 100],
                ["2019-09-02", "A", 300],
                ["2019-09-03", "A", 10],
            ],
        )
        df4 = pd.DataFrame(
            columns=columns2,
            data=[
                ["2019-09-01", "B", 68],
                ["2019-09-02", "B", 35],
                ["2019-09-03", "B", 30],
            ],
        )
        dfs = [df1, df2, df3, df4]
        self._test_slices("letter", ["opportunities", "impressions"], 4, dfs)

    def test_slice_bad_dimension(self):
        """Tests the slice on a bad dimension."""
        with self.assertRaises(ValueError):
            self._test_slices("lette", "opportunities", 0, 0)


if __name__ == "__main__":
    main()



"""Testing the database case."""
import logging
import os
from unittest import TestCase

from muttlib.dbconn import get_client_from_connstr
from sqlalchemy.exc import ResourceClosedError

logger = logging.getLogger(__name__)

TEST_DB_CONNSTR = "TEST_DB_CONNSTR"
TEST_DB_NAME = "pytests"
POSTGRES_DB = "soam_db"


class PgTestCase(TestCase):
    """Creates the Test Case object."""

    db_connstr = os.getenv(TEST_DB_CONNSTR)
    if db_connstr:
        _, db_client = get_client_from_connstr(db_connstr)
        prev_database = db_client.database
        # db_client.database = TEST_DB_NAME
    else:
        logger.warning(f"Missing test database connection string in {TEST_DB_CONNSTR}")

    @classmethod
    def run_query(cls, query):
        cursor = cls.db_client.execute(query)
        try:
            ret = list(cursor)
        except ResourceClosedError:  # no results to fetch
            ret = None
        cursor.close()
        return ret

    def setUp(self):
        # self.db_client.database =  os.getenv(POSTGRES_DB)
        self.db_client._engine = None  # pylint:disable=protected-access
        conn = self.db_client._connect()  # pylint:disable=protected-access
        # DROP/CREATE DATABASE fails in a normal session.
        # Ref: https://stackoverflow.com/questions/44959599/sql-alchemy-cannot-run-inside-a-transaction-block
        query = f"""
        DROP DATABASE IF EXISTS {TEST_DB_NAME};
        """
        conn.execution_options(isolation_level="AUTOCOMMIT").execute(query)
        query = f"""
        CREATE DATABASE {TEST_DB_NAME};
        """
        conn.execution_options(isolation_level="AUTOCOMMIT").execute(query)
        conn.close()
        self.db_client._engine = None  # pylint:disable=protected-access
        # self.db_client.database = TEST_DB_NAME

    def tearDown(self):
        # FIXME: This should drop the test database but it fails. The reason seem to be
        # that seems that one or more connection are leaking during the tests.
        # del self.db_client
        # self.db_client = None
        # db_connstr = os.getenv(TEST_DB_CONNSTR)
        # _, db_client = get_client_from_connstr(self.db_connstr)
        # conn = db_client._connect()
        # query = f"DROP DATABASE IF EXISTS {TEST_DB_NAME}"
        # conn.execution_options(isolation_level="AUTOCOMMIT").execute(query)
        # conn.close()
        pass



"""statsmodels.SARIMAX estimators."""
import logging
from typing import Dict, List, Tuple

import pandas as pd

from soam.constants import DS_COL, YHAT_COL
from soam.models.base import SkWrapper, sk_constructor_wrapper

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

try:
    from statsmodels.tsa.statespace.sarimax import SARIMAX
except ImportError:
    logger.warning("No SARIMAX support")
    logger.warning("If you want to use it, ´pip install soam[statsmodels]´")


class SkSarimax(SkWrapper):
    """Scikit-Learn statsmodels.SARIMAX model wrapper."""

    @sk_constructor_wrapper(SARIMAX)
    def __init__(  # pylint: disable=super-init-not-called
        self,
        extra_regressors: List[str] = None,
        fit_params: Dict = None,
        date_col: str = DS_COL,
    ):
        """Construct wrapper with extra parameters in addition to the SARIMAX ones.

        Parameters
        ----------
        extra_regressors : List[str], optional
            Column names of exogenous variables, by default None
        fit_params : Dict, optional
            Params to be fed to model's .fit() method, by default None
        date_col : str, optional
            date column, by default None
        """
        self.extra_regressors = extra_regressors
        self.fit_params = fit_params or {}
        self.date_col = date_col

    def fit(self, X: pd.DataFrame, y: pd.Series):
        """
        Fit estimator to data.

        Notes:
            Since SARIMAX requires endog and exog arrays (not colnames) at
            init time, we transform the input before _init_sk_model.
        """
        # arrays are required at model initialization

        (
            self.exog,  # pylint: disable=attribute-defined-outside-init
            self.endog,  # pylint: disable=attribute-defined-outside-init
        ) = self._transform_to_input_format(X, y)
        self.model = self._init_sk_model(  # pylint: disable=attribute-defined-outside-init
            SARIMAX, clean=True
        )

        self.model_fit = self.model.fit(  # pylint: disable=attribute-defined-outside-init
            **self.fit_params or {}
        )
        # saving as prediction start point
        self._train_len = len(X)  # pylint: disable=attribute-defined-outside-init
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's predict."""
        exog_predict, _ = self._transform_to_input_format(X)
        start = self._train_len
        end = start + X.pipe(len) - 1
        predictions = self.model_fit.predict(start, end, exog=exog_predict)
        predictions = self._transform_to_output_format(predictions, X)
        return predictions.reset_index(drop=True)

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """Scikit learn's transform."""
        return self.predict(X)

    def fit_transform(self, X: pd.DataFrame, y: pd.Series, output_length: int = 1):
        """Scikit learn's fit_transform with output_length."""
        X_train = X[:-output_length]
        X_pred = X[-output_length:]
        y_train = y[:-output_length]
        self.fit(X_train, y_train)
        return self.transform(X_pred)

    def _transform_to_input_format(
        self, X: pd.DataFrame, y: pd.Series = None
    ) -> Tuple[pd.DataFrame, pd.Series]:
        """Transform input to SARIMAX compatible format.

        Notes:
            Since SARIMAX only takes arrays as x,y inputs,
            we select the extra_regressors cols from X.
        """

        if self.extra_regressors is not None:
            return X[self.extra_regressors], y
        else:
            return None, y

    def _transform_to_output_format(
        self, predictions: pd.Series, X_pred: pd.DataFrame
    ) -> pd.DataFrame:
        """Transform SARIMAX output to SoaM format."""
        final_predictions = pd.DataFrame()
        final_predictions[self.date_col] = X_pred[self.date_col].values
        final_predictions[YHAT_COL] = predictions.values

        return final_predictions



"""Transformer tester."""
from unittest import TestCase

import pandas as pd
from sklearn.preprocessing import StandardScaler

from soam.workflow import BaseDataFrameTransformer, Transformer


class SimpleProcessor(BaseDataFrameTransformer):
    """Creates the Simple Processor Object."""

    def __init__(self, **fit_params):
        self.preproc = StandardScaler(**fit_params)

    def fit(self, df_X):
        """Fit."""
        self.preproc.fit(df_X['a'].values.reshape(-1, 1))
        return self

    def transform(self, df_X, inplace=True):
        """Transform."""
        if not inplace:
            df_X = df_X.copy()
        df_X['a'] = self.preproc.transform(df_X['a'].values.reshape(-1, 1))
        return df_X


class TransformerTestCase(TestCase):
    """Creates the Transformer Test Case Object."""

    def test_simple_case(self):
        """Simple case testing."""
        test_data_X = pd.DataFrame({'a': [1, 2, 3]})
        test_data_X2 = pd.DataFrame({'a': [4, 6, 8]})
        expected_output = pd.DataFrame({"a": [-1.2247448714, 0.0, 1.2247448714]})
        expected_output_2 = pd.DataFrame({"a": [2.449490, 4.898979, 7.348469]})

        preproc = Transformer(SimpleProcessor())
        transformed_dataset, _ = preproc.run(test_data_X)

        pd.testing.assert_frame_equal(transformed_dataset, expected_output)
        pd.testing.assert_frame_equal(
            preproc.transform(test_data_X2), expected_output_2
        )



# runner.py
"""
SoamFlow Class
--------------
A class that can execute the pipeline steps and keep track of the whole run
data

TODO: review if SoamFlow is going to be the only class that can run the
 pipeline
"""
from datetime import datetime
import logging
from typing import TYPE_CHECKING, Optional  # pylint:disable=unused-import

from prefect import Flow, Task

from soam.cfg import TRACKING_IS_ACTIVE, TRACKING_URI

logger = logging.getLogger(__name__)
try:
    from mlflow import end_run, set_tracking_uri, start_run
except ModuleNotFoundError:
    logger.debug("Mlflow dependency is not installed.")


if TYPE_CHECKING:
    from soam.savers.savers import Saver


class SoamFlow(Flow):
    """
    Soam Flow to execute the pipeline steps and keep track of the whole run data.
    SoamFlow is an extension of prefect.Flow to add tracking functionality.
    """

    def __init__(self, saver: "Optional[Saver]" = None, **kwargs):
        """
        Soam Flow init to execute pipeline steps and keep track of the run data.

        Parameters
        ----------
        saver: soam.savers.Saver
            The saver to store the pipeline steps and keep track of the whole run data.
        kwargs: dict
            extra args.
        """

        super().__init__(**kwargs)
        self.saver = saver
        self.start_datetime: datetime
        self.end_datetime: datetime
        if self.saver is not None:
            self.state_handlers.append(self.saver.save_flow_run)

        if TRACKING_IS_ACTIVE:
            self.active_run = None
            set_tracking_uri(TRACKING_URI)
            self.state_handlers.append(self.set_tracker_run)

    def add_task(self, task: Task) -> Task:
        """
        Adds the task and handlers to save the state.

        Parameters
        ----------
        task
            Represents a task where the flow is being executed.

        Returns
        -------
        super object
            Saves the task and the state handler.
        """
        if self.saver is not None:
            if self.saver.save_task_run not in task.state_handlers:
                task.state_handlers.append(self.saver.save_task_run)
        return super().add_task(task)

    def set_tracker_run(
        self, obj, old_state, new_state
    ):  # pylint: disable=unused-argument
        """
        Create a new mlflow run.

        Parameters
        ----------
        obj
            the underlying object to which this state handler is attached
        old_state
            the previous state of this object
        new_state
            the proposed new state of this object
        Returns
        -------
        newstate
            the new state of this object.
        """
        if new_state.is_running():
            obj.active_run = start_run(run_name="flow_run")
        if new_state.is_finished():
            end_run()
        return new_state



"""Tests for TimeSeriesExtractor.
"""
import os
import unittest
from unittest import main

from muttlib.dbconn import get_client_from_connstr
import pandas as pd
from sqlalchemy import Column
from sqlalchemy.types import Integer, String

from soam.data_models import AbstractIDBase
from soam.workflow import Store
from tests.db_test_case import TEST_DB_CONNSTR, PgTestCase

TABLE_STORE = "test_store"


class ConcreteTableStore(AbstractIDBase):
    __tablename__ = TABLE_STORE

    country = Column(String(3))
    metric1 = Column(Integer())


@unittest.skipIf(not os.getenv(TEST_DB_CONNSTR), f"{TEST_DB_CONNSTR} is not set")
class TestStore(PgTestCase):
    def test_store_df(self):
        """Test storing data and retriving it."""
        df = pd.DataFrame(
            {
                "id": [1, 2, 3],
                "country": ["ARG", "BRZ", "ARG"],
                "metric1": [328, 215, 146],
            }
        )
        self.store.run(df)
        sql_select = f"""
        SELECT * FROM {TABLE_STORE}
        """
        db_connstr = os.getenv(TEST_DB_CONNSTR)
        _, db_client = get_client_from_connstr(db_connstr)
        pd.testing.assert_frame_equal(db_client.to_frame(sql_select), df)

    @classmethod
    def setUpClass(cls):
        super().setUp(cls)

        db_connstr = os.getenv(TEST_DB_CONNSTR)
        engine = cls.db_client.get_engine(db_connstr)
        ConcreteTableStore.__table__.create(engine)
        cls.store = Store(cls.db_client, f"{TABLE_STORE}")

    @classmethod
    def tearDownClass(cls):
        super().tearDown(cls)
        db_connstr = os.getenv(TEST_DB_CONNSTR)
        engine = cls.db_client.get_engine(db_connstr)
        ConcreteTableStore.__table__.drop(engine)

    def setUp(self):
        pass

    def tearDown(self):
        pass


if __name__ == "__main__":
    main()



# slack_report.py
"""
Slack Report
------------
Slack reporting and message formatting tools. Its a postprocess that sends the
model forecasts though the slack app.
"""
from asyncio import Future
from io import BytesIO
from pathlib import Path, PosixPath
from typing import IO, Any, Dict, Optional, Union

from jinja2 import Template
from muttlib.utils import path_or_string
import pandas as pd
import slack
from slack.web.slack_response import SlackResponse

from soam.cfg import get_slack_cred
from soam.constants import DS_COL, YHAT_COL
from soam.core.step import Step

DEFAULT_GREETING_MESSAGE = "Hello everyone! Here are the results of the forecast for the *{metric_name}* metric:\n"
DEFAULT_FAREWELL_MESSAGE = "Cheers!\n SoaM."


class SlackReport:
    """
    Generates the report to share via Slack.
    """

    def __init__(
        self, channel_id: str, metric_name: str, setting_path: Optional[str],
    ):
        """
        Initialization of the Slack Report object.

        Parameters
        ----------
        channel_id: str
            Slack channel id where the report will be sent.
        metric_name: str
            Performance metric being measured.
        setting_path: str
            Setting path.
        """
        credentials = get_slack_cred(setting_path)
        self.slack_client = slack.WebClient(credentials["token"])
        self.channel_id = channel_id
        self.metric_name = metric_name

    def send_report(
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ) -> Union[Future, SlackResponse]:
        """
        Send Slack report.

        Parameters
        ----------
        prediction : pd.DataFrame
            DataDrame of the predictions made.
        plot_filename : str or pathlib.Path
             Path of the forecast data to send.
        greeting_message: str
            Greeting message to send via Slack with the predictions.
        farewell_message: str
            Farewell message to send via Slack with the predictions.

        Returns
        -------
        Slack Report
            Sends the specified message with the predictions data via Slack.
        """
        if greeting_message == DEFAULT_GREETING_MESSAGE:
            greeting_message.format(metric_name=self.metric_name)

        summary_entries = []
        summary_entries.append(greeting_message)

        for _, row in prediction.iterrows():
            date = row[DS_COL].strftime('%Y-%b-%d')
            value = "{:.2f}".format(row[YHAT_COL])
            summary_entries.append(f"• *[{date}]* {value}\n")

        summary_entries.append(farewell_message)

        summary_message = "\n".join(summary_entries)  # type: ignore

        return self.slack_client.files_upload(
            channels=self.channel_id,
            file=str(plot_filename),
            initial_comment=summary_message,
            title=f"{self.metric_name} Forecast",
        )


class SlackReportTask(Step, SlackReport):
    """
    Builds up the task of the report designed for Slack.
    """

    def __init__(
        self,
        channel_id: str,
        metric_name: str,
        setting_path: Optional[str],
        **kwargs: Any,
    ):
        """
        Parameters
        ----------
        channel_id: str
            Slack channel id where the report will be sent.
        metric_name: str
            Performance metric being measured.
        setting_path: str
            Setting path.
        kwargs:
            Extra args to pass.
        """
        Step.__init__(self, **kwargs)  # type: ignore
        SlackReport.__init__(
            self, channel_id, metric_name, setting_path,
        )

    def run(  # type: ignore
        self,
        prediction: pd.DataFrame,
        plot_filename: Union[str, Path],
        greeting_message: Optional[str] = DEFAULT_GREETING_MESSAGE,
        farewell_message: Optional[str] = DEFAULT_FAREWELL_MESSAGE,
    ):
        """
        Send Slack report task.

        Parameters
        ----------
        prediction : pd.DataFrame
            DataDrame of the predictions made.
        plot_filename : str or pathlib.Path
             Path of the forecast data to send.
        greeting_message: str
            Greeting message to send via Slack with the predictions.
        farewell_message: str
            Farewell message to send via Slack with the predictions.

        Returns
        -------
        Slack Report Task
            Sends the specified message with the predictions data via Slack.
        """
        return self.send_report(
            prediction, plot_filename, greeting_message, farewell_message
        )


class SlackMessage:
    def __init__(
        self,
        template,
        arguments: Dict = None,
        attachment: Optional[Union[Path, IO]] = None,
    ):
        """Create Slack message.

        Args:
            template: Jinja template.
            template_args (dict): Arguments to be passed to the Jinja templates .
            attachment (Path): Path to file (or image) to attach to the image.
        """
        self.template = template
        self.arguments = arguments
        self.attachment_ref = attachment

    @property
    def message(self) -> str:
        """Message property."""
        message = path_or_string(self.template)
        if self.arguments:
            template = Template(message)
            message = template.render(**self.arguments)  # type:ignore
        return message

    @property
    def attachment(self) -> Optional[Union[str, IO]]:
        """Message property."""
        if self.attachment_ref is None:
            return None
        if isinstance(self.attachment_ref, PosixPath):
            if not self.attachment_ref.exists():
                raise ValueError(
                    f"File does not exist: {str(self.attachment_ref.resolve())}."
                )
            # slack's client supports a string with the path for the file
            return str(self.attachment_ref.resolve())
        elif isinstance(self.attachment_ref, BytesIO):
            return self.attachment_ref
        else:
            raise TypeError("Only PosixPath and BytesIO supported.")


def send_slack_message(
    slack_client: slack.WebClient,
    channel: str,
    msg: SlackMessage,
    thread_ts: Optional[int] = None,
):
    """Send Slack message.

    Parameters
    ----------
    channel : str
        slack channel to send the message to.
    msg : SlackMessage
        SlackMessage instance.
    thread_ts : int, optional
        message timestamp to reply to in threaded fashion, by default None.
    """
    if msg.attachment is None:
        response = slack_client.chat_postMessage(
            channel=channel, text=msg.message, thread_ts=thread_ts
        )
    else:
        response = slack_client.files_upload(
            file=msg.attachment,
            channels=channel,
            initial_comment=msg.message,
            thread_ts=thread_ts,
        )
    return response



"""CSV saver."""
from pathlib import Path
from typing import Union

from filelock import FileLock
from muttlib.utils import make_dirs
import pandas as pd
from prefect import Task, context
from prefect.engine.state import State

from soam.constants import FLOW_FILE_NAME, LOCK_NAME
from soam.core import SoamFlow
from soam.savers.savers import Saver
from soam.utilities.utils import get_file_path


class CSVSaver(Saver):
    """
    CSV Saver object to store the predictions and the runs.
    """

    def __init__(self, path: Union[str, Path]):
        """
        Create a saver object to store the predcitions and the runs.

        You can use a CSVSaver for storing runs and/or values.

        Parameters
        ----------
        path
            str or pathlib.Path where the file will be created.
        """
        super().__init__()
        self.path = Path(make_dirs(path))

    @property
    def flow_path(self) -> Path:
        """
        Generation of the flow run folder directory.

        Parameters
        ----------
        Path
            str where the file will be created.

        Returns
        -------
        path
            Path of the directory of the flow run folder.
        """
        flow_run_folder = (
            context["flow_name"]
            + "_"
            + context["date"].replace(microsecond=0, tzinfo=None).isoformat()
            + "_"
            + context["flow_run_id"]
        )

        return make_dirs(self.path / flow_run_folder)

    @property
    def flow_file_path(self) -> Path:
        """
        Flow file name upon the path.

        Parameters
        ----------
        Path
            str where the file will be created.

        Returns
        -------
        path
            Path of the flow file.
        """
        return self.flow_path / FLOW_FILE_NAME

    @property
    def flow_run_lock(self) -> Path:
        """
        Flow run lock upon the path.

        Parameters
        ----------
        Path
            str where the file will be created.

        Returns
        -------
        path
            Path of the lock name.
        """
        return self.flow_path / LOCK_NAME

    def save_forecast(self, task: Task, old_state: State, new_state: State) -> State:
        """
        Store the forecaster data in the constructed path
        with the `{task_slug}_forecasts.csv`.

        Parameters
        ----------
        task: Task
            Specify the forecast task you want to save information of.
        old_state: State
            Task old state.
        new_state : State
            Task new state.

        Returns
        -------
        State
            The new updated state of the forecast task.
        """
        if new_state.is_successful():
            save_prediction = new_state.result[0].copy()
            save_prediction["task_run_id"] = context["task_run_id"]

            task_run_id = context["task_slug"] + "_" + context["task_run_id"]
            PREDICTION_CSV_SUFFIX = f"{task_run_id}_forecasts.csv"
            prediction_path = get_file_path(self.flow_path, PREDICTION_CSV_SUFFIX)

            save_prediction.to_csv(prediction_path, index=False)

        return new_state

    def save_task_run(self, task: Task, old_state: State, new_state: State) -> State:
        """
        Store the task run information in the csv file created by `save_flow_run`

        Parameters
        ----------
        task: Task
            Specify the task you want to save information of.
        old_state: State
            Task old state.
        new_state : State
            Task new state.

        Returns
        -------
        State
            The new updated state of the task.
        """
        if new_state.is_successful():
            flow_run_file = self.flow_file_path
            lock = FileLock(str(self.flow_run_lock))

            with lock.acquire(timeout=5):
                read_df = pd.read_csv(flow_run_file)
                flow_values = read_df[read_df.flow_state.str.contains("Running")].iloc[
                    0
                ]
                csv_data = {
                    "flow_run_id": [flow_values["flow_run_id"]],
                    "start_datetime": [flow_values["start_datetime"]],
                    "end_datetime": [flow_values["end_datetime"]],
                    "flow_state": [flow_values["flow_state"]],
                    "task_run_id": [context["task_run_id"]],
                    "repr_task": [repr(task)],
                }

                df = pd.DataFrame.from_dict(csv_data)
                df.to_csv(flow_run_file, mode="a", header=False, index=False)

        return new_state

    def save_flow_run(
        self, soamflow: SoamFlow, old_state: State, new_state: State
    ) -> State:
        """
        Store the SoamFlow run information, create a folder for the run with a csv file.

        Parameters
        ----------
        saomflow: SoamFlow
            Specify the soamflow you want to save information of.
        old_state: State
            SoamFlow old state.
        new_state : State
            SoamFlow new state.

        Returns
        -------
        State
            The new updated state of the SoamFlow.
        """
        if new_state.is_running():
            csv_data = {
                "flow_run_id": [context["flow_run_id"]],
                "start_datetime": [soamflow.start_datetime],
                "end_datetime": [soamflow.start_datetime],
                "flow_state": [new_state.message],
                "task_run_id": None,
                "repr_task": None,
            }

            df = pd.DataFrame.from_dict(csv_data)
            df.to_csv(self.flow_file_path, index=False)

        elif new_state.is_successful() or new_state.is_failed():
            self.flow_run_lock.unlink()

        return new_state



from unittest.mock import MagicMock, patch

import pandas as pd
from pandas.testing import assert_frame_equal
import pytest

from soam.constants import DS_COL, YHAT_COL
from soam.models.sarimax import SkSarimax
from soam.utilities.utils import add_future_dates
from tests.helpers import sample_data_df  # pylint: disable=unused-import


def test_fit(sample_data_df):  # pylint: disable=redefined-outer-name
    with patch("soam.models.sarimax.SARIMAX") as model_patch:
        output_length = 10
        train_data = sample_data_df
        data = add_future_dates(train_data, output_length)
        X, y = data[data.columns[:-1]], data[data.columns[-1]]
        wrapper = SkSarimax()
        model_patch.assert_not_called()
        wrapper.fit(X, y)
        model_patch.assert_called_once()
        model_patch().fit.assert_called_once()


def test_predict_without_fit_fails(
    sample_data_df,
):  # pylint: disable=redefined-outer-name
    with patch("soam.models.sarimax.SARIMAX") as model_patch:
        output_length = 10
        train_data = sample_data_df
        data = add_future_dates(train_data, output_length)
        X = data[data.columns[:-1]]
        wrapper = SkSarimax()
        model_patch.assert_not_called()
        with pytest.raises(AttributeError):
            wrapper.predict(X)


def test_fit_transform(sample_data_df,):  # pylint: disable=redefined-outer-name
    with patch("soam.models.sarimax.SARIMAX") as model_patch:
        output_length = 10
        train_data = sample_data_df
        data = add_future_dates(train_data, output_length)
        X, y = data[data.columns[:-1]], data[data.columns[-1]]
        wrapper = SkSarimax()
        model_patch.assert_not_called()
        # Setup mock for fit() return value
        pred_mock = MagicMock()
        prediction_values = pd.Series([i for i in range(output_length)])
        pred_mock.predict.return_value = prediction_values
        model_patch.return_value.fit.return_value = pred_mock
        # Run fit_transform
        predictions = wrapper.fit_transform(X, y, output_length=output_length)
        model_patch.assert_called_once()
        model_patch.return_value.fit.assert_called_once()
        pred_mock.predict.assert_called_once()
        # Setup expected prediction
        expected_predictions = pd.DataFrame(
            {
                DS_COL: X[DS_COL].values[-output_length:],
                YHAT_COL: prediction_values.values,
            }
        )
        assert_frame_equal(predictions, expected_predictions)
